==> tecfy_core_package-1.1.27.tar.gz: lib/controllers/entity/entity_list.controller.dart <==
part of tecfy_core_client_package;

class EntityListController extends BaseController {
  late Entity entity;
  bool isPopup;

  /// data
  List<dynamic>? data;
  List<dynamic>? filterdData;
  List<EntityFilter>? filters;

  late List<Property> searchableProperties;
  int pageIndex = 0;
  int? totalCount;
  int pageIndexMax = 0;
  bool hasNextPage = true;
  late ScrollController scrollController;
  late TextEditingController searchController;
  late RefreshController refreshController =
      RefreshController(initialRefresh: false); // for pull to refrsh
  IEntityAction? entityActionsService;
  EntityListController(State<StatefulWidget> state, String entityName,
      {this.filters,
      this.orderBy,
      this.orderByDesc = false,
      required this.isPopup})
      : super(state) {
    scrollController = ScrollController();
    refreshController.footerMode?.addListener(footerListner);
    searchController = TextEditingController();
    if (filters == null) filters = []; // List<EntityFilter>();
    if (kDebugMode) print('-------------- EntityListController => $entityName');
    entity = EntityService.getEntity(entityName) ?? Entity.fromMap({});
    searchableProperties =
        entity.properties.where((e) => e.isSearchable).toList();

    entityActionsService =
        TecfyCoreApp.entityActionsServices[entity.dbName.toLowerCase()];

    loadData();
  }

  @override
  dispose() {
    try {
      refreshController.dispose();
    } catch (err) {}
    try {
      searchController.dispose();
    } catch (err) {}
    try {
      scrollController.dispose();
    } catch (err) {}
    super.dispose();
  }

  String? orderBy;
  bool orderByDesc = true;

  addChildren(items, List<dynamic> rows) {
    for (var item in items) {
      rows.add(item);
      if (item['_expanded'] == true &&
          item['_children'] != null &&
          item['_children'].length > 0) {
        rows = addChildren(item['_children'], rows);
      }
    }
    return rows;
  }

  Future<void> onOpenFilter() async {
    var filterResult = await AppNavigator.showAppDialog(
      context,
      (c) => EntityFilterPage(entity.properties, filters),
    );
    if (kDebugMode) print('-------------------- dialog result $filterResult');
    if (filterResult != null && filterResult is List<EntityFilter>) {
      filterResult = filterResult
          .where((e) =>
              e.value != null && (e.propertyName != null || e.property != null))
          .toList();
      if (filterResult.length != 0) {
        filters = filterResult;
        loadData();
      }
    }
  }

  Future<void> onOpenMerge() async {
    var mergeResult = await AppNavigator.showAppDialog(
        context, (c) => EntityMergePage(entity: entity));
    if (mergeResult == true) {
      loadData();
    }
  }

  Future globalSearchText() async {
    if (ConnectionService.isOfflineMode && searchController.text.isNotEmpty) {
      globalSearchTextServer();
      return;
    }
    if (searchController.text.isEmpty) {
      filterdData = data;
    } else {
      filterdData = data?.where((d) {
        return searchableProperties
            .any((p) => d[p.dbName].toString().contains(searchController.text));
      }).toList();
    }

    if (entity.isTree) {
      filterdData = addChildren(filterdData, List.empty(growable: true));
    }
    setState();
  }

  Future globalSearchTextServer() async {
    if (searchController.text.isEmpty) {
      //filterdData = data;
      await loadFromServer();
      return;
    }
    if (kDebugMode) print('start search server -> ${searchController.text}');
    var result = await EntityService.getDataSearch(
      entity,
      searchController.text,
      pageIndex: pageIndex,
    );
    if (kDebugMode) print(result);
    if (!result.success) {
      result.showMessage(context);
    }

    if (result.data != null) {
      if (kDebugMode)
        print('-------------- ${result.data?.length}  records loaded ');
      filterdData = result.data;
    } else {
      filterdData = [];
    }
    if (entity.isTree) {
      filterdData = addChildren(filterdData, List.empty(growable: true));
    }
    setState();
  }

  sortBy(Property prop) async {
    if (kDebugMode) print('sort by ${prop.dbName}');
    if (prop.dbName != orderBy) {
      orderBy = prop.dbName;
      orderByDesc = false;
    } else {
      orderByDesc = !orderByDesc;
    }
    setLoading();
    await globalSearchTextServer();
    setResume();
  }

  bool _hasPageing = false;
  bool get hasPageing {
    if ((data?.length ?? 0) >= EntityService.pageSize) _hasPageing = true;
    return _hasPageing;
  }

  footerListner() async {
    if (kDebugMode)
      print('==========on load ====== ${refreshController.footerStatus}');
    if (refreshController.footerStatus != LoadStatus.loading) return;
    var success = await gotoMextPageIndex();
    if (success && hasNextPage) {
      refreshController.loadComplete();
    } else if (success && !hasNextPage) {
      //refreshController.loadComplete();
      refreshController.loadNoData();
    } else {
      refreshController.loadFailed();
    }
  }

  Future<bool> gotoMextPageIndex() {
    pageIndex++;
    pageIndexMax = max(pageIndexMax, pageIndex);
    return loadData(withLoading: false);
  }

  gotoPageIndex(int i) {
    pageIndex = i;
    setLoading();
    loadData();
  }

  Future getTotalCount() async {
    setLoading();
    if (searchController.text.isEmpty) {
      var result = await EntityService.getCount(entity,
          filter: filters ?? [], filterParentId: entity.isTree);
      if (result.data != null) {
        totalCount = result.data;
      }
    } else {
      List<EntityFilter> f = [];
      for (var prop in entity.properties.where((p) => p.isSearchable)) {
        f.add(EntityFilter(
            propertyName: prop.dbName,
            operatorId: prop.valueType == ValueTypes.Number ? 'E' : 'C',
            value: searchController.text));
      }
      var d = (await EntityService.getCount(entity,
              filter: f, filterOperator: 'or'))
          .data;
      if (d != null) totalCount = d;
    }
    setResume();
    setState();
  }

  var isLoading = false;
  Future<bool> loadData({bool withLoading = true}) async {
    if (isLoading) return true;
    isLoading = true;
    var result = await EntityService.getData(entity,
        orderBy: orderBy,
        orderByDesc: orderByDesc,
        filter: filters,
        pageIndex: pageIndex,
        filterParentId: entity.isTree);
    if (withLoading) setResume(result.success);
    if (!result.success || result.data == null) {
      //data = [];
      isLoading = false;
      if (withLoading) {
        setState();
      }
      if (ConnectionService.isOnlineMode) {
        showMessage(
          'Error Loading From Server',
          MessageType.error,
          message: result.message,
        );
      } else {
        showMessage('Error Loading From Local Database', MessageType.error);
      }
      globalSearchText();
      return false;
    } else {
      if (LayoutService.isMobile(context) && pageIndex > 0 && !entity.isLocal) {
        data?.addAll(result.data ?? []);
      } else {
        data = result.data;
      }
      if ((result.data?.length ?? 0) < EntityService.pageSize)
        hasNextPage = false;
    }
    globalSearchText();
    // if(pageIndex == 0) refreshAnimation();
    isLoading = false;
    return true;
  }

  Future loadFromServer() async {
    await loadData();

    refreshController.refreshCompleted();
  }

  Future onRefresh() async {
    pageIndex = 0;
    pageIndexMax = 0;
    hasNextPage = true;
    setLoading();
    await loadData();
  }

  Future onAdd() async {
    if (isPopup) {
      var result = await AppNavigator.showAppDialogEx(
          context, () => EntityFormPage(entity.dbName, isPopup: true),
          animationType: LayoutService.isMobile(context)
              ? AnimationTypes.slideUp
              : AnimationTypes.opacity,
          size: DialogBoxSizes.large);
      if (result == true) {
        loadFromServer();
      }
      return;
    }
    if (entityActionsService?.hasAddPage() ?? false) {
      await entityActionsService!.addPage(context);
      //if (result == true) {
      loadFromServer();
      // }
    } else {
      var result = await AppNavigator.navigateTo(
          context, () => EntityFormPage(entity.dbName),
          animationType: LayoutService.isMobile(context)
              ? AnimationTypes.slideUp
              : AnimationTypes.opacity);
      if (result == true) {
        loadFromServer();
      }
    }
  }

  Future onUpdate(d) async {
    if (isPopup) {
      var result = await AppNavigator.showAppDialogEx(
          context, () => EntityFormPage(entity.dbName, data: d, isPopup: true),
          animationType: LayoutService.isMobile(context)
              ? AnimationTypes.slideUp
              : AnimationTypes.opacity,
          size: DialogBoxSizes.large);
      if (result == true) {
        loadFromServer();
      }
      return;
    }
    if (entityActionsService?.hasEditPage(d) ?? false) {
      await entityActionsService!.editPage(context, d);
      loadFromServer();
    } else {
      var result = await AppNavigator.navigateTo(
          context, () => EntityFormPage(entity.dbName, data: d),
          animationType: LayoutService.isMobile(context)
              ? AnimationTypes.slideUp
              : AnimationTypes.opacity);
      if (result == true) {
        loadFromServer();
      }
    }
  }

  Future onView(d) async {
    if (entityActionsService?.hasEditPage(d) ?? false) {
      await entityActionsService!.editPage(context, d);
      loadFromServer();
    } else {
      var result = await AppNavigator.navigateTo(
          context, () => EntityFormPage(entity.dbName, data: d),
          animationType: LayoutService.isMobile(context)
              ? AnimationTypes.slideUp
              : AnimationTypes.opacity);
      if (result == true) {
        loadFromServer();
      }
    }
  }

  Future onDelete(d) async {
    if (!entity.permission.canDelete ||
        entity.entityType == EntityTypes.ViewOnly ||
        entity.entityType == EntityTypes.Lookup) {
      return showMessage('لا يوجد صلاحيه للحذف', MessageType.error);
    }
    if (await confirmMessage('هل انت متاكد من الحذف؟')) {
      var result = await EntityService().delete(entity, d['id']);
      if (result.success) {
        showMessage('تم الحذف بنجاح', MessageType.success);
        data?.removeWhere((e) => e['id'] == d['id']);
        globalSearchText();
        //loadFromServer();
      } else {
        showMessage(
          'خطأ اثناء الحذف',
          MessageType.error,
          message: 'الرجاء التاكد من ان العنصر غير مرتبط بعناصر اخرى',
        );
      }
    }
  }

  Future expandItem(Map<String, dynamic> d) async {
    try {
      if (!entity.isTree) return true;
      if (d['_children'] != null) {
        d['_expanded'] = !d['_expanded'];
        globalSearchText();
        return;
      }

      if (isLoading) return true;
      isLoading = true;
      if (kDebugMode) print('expand ${d['id']}');
      setLoading();
      var result = await EntityService.getData(entity,
          filter: filters,
          pageIndex: pageIndex,
          filterParentId: true,
          orderBy: orderBy,
          orderByDesc: orderByDesc,
          parentId: d['id'].toString());

      setResume();
      isLoading = false;
      if (!result.success) {
        //data = [];
        return;
      }
      for (var r in result.data ?? []) {
        if (d['_tree_level'] != null)
          r['_tree_level'] = d['_tree_level'] + 1;
        else
          r['_tree_level'] = 1;
      }
      d['_children'] = result.data;
      d['_expanded'] = true;
      // data.insertAll(
      //     data.indexWhere((i) => i['id'] == d['id']) + 1, result.data);
      globalSearchText();
    } catch (err, t) {
      print('XXX expand item error');
      print(err);
      if (kDebugMode) print(t);
      setResume();
      isLoading = false;
    }
  }

  bool hasActions(Map<String, dynamic>? d) {
    if (entity.actions.length > 0) return true;
    if (entityActionsService == null) return false;
    return (entityActionsService?.getActions(d).data?.length ?? 0) > 0;
  }

  List<EAction> getActions(Map<String, dynamic>? d) {
    return EntityService.getActions(entity, d);
  }

  Future onAction(EAction action, Map<String, dynamic> data) async {
    try {
      return await EntityService.onAction(entity, action, data, context);
    } catch (err) {
      showMessage('خطأ اثناء تنفيذ الاجراء', MessageType.error);
    }
  }

  lowerFirst(String s) {
    return '${s[0].toLowerCase()}${s.substring(1)}';
  }

  String get getPhotoDataPropName =>
      '${entity.properties.firstWhere((d) => d.isImage).dbName}Id';
}

==> tecfy_core_package-1.1.38.tar.gz: lib/controllers/entity/entity_list.controller.dart <==
part of tecfy_core_client_package;

class EntityListController extends BaseController {
  late Entity entity;
  bool isPopup;

  /// data
  List<dynamic>? data;
  List<dynamic>? filterdData;
  List<EntityFilter>? filters;

  late List<Property> searchableProperties;
  int pageIndex = 0;
  int? totalCount;
  int pageIndexMax = 0;
  bool hasNextPage = true;
  late ScrollController scrollController;
  late TextEditingController searchController;
  late RefreshController refreshController =
      RefreshController(initialRefresh: false); // for pull to refrsh
  IEntityAction? entityActionsService;
  EntityListController(State<StatefulWidget> state, String entityName,
      {this.filters,
      this.orderBy,
      this.orderByDesc = false,
      required this.isPopup})
      : super(state) {
    scrollController = ScrollController();
    refreshController.footerMode?.addListener(footerListner);
    searchController = TextEditingController();
    if (filters == null) filters = []; // List<EntityFilter>();
    if (kDebugMode) print('-------------- EntityListController => $entityName');
    entity = EntityService.getEntity(entityName) ?? Entity.fromMap({});
    searchableProperties =
        entity.properties.where((e) => e.isSearchable).toList();

    entityActionsService =
        TecfyCoreApp.entityActionsServices[entity.dbName.toLowerCase()];

    loadData();
  }

  @override
  dispose() {
    try {
      refreshController.dispose();
    } catch (err) {}
    try {
      searchController.dispose();
    } catch (err) {}
    try {
      scrollController.dispose();
    } catch (err) {}
    super.dispose();
  }

  String? orderBy;
  bool orderByDesc = true;

  addChildren(items, List<dynamic> rows) {
    for (var item in items) {
      rows.add(item);
      if (item['_expanded'] == true &&
          item['_children'] != null &&
          item['_children'].length > 0) {
        rows = addChildren(item['_children'], rows);
      }
    }
    return rows;
  }

  Future<void> onOpenFilter() async {
    var filterResult = await AppNavigator.showAppDialog(
      context,
      (c) => EntityFilterPage(entity.properties, filters),
    );
    if (kDebugMode) print('-------------------- dialog result $filterResult');
    if (filterResult != null && filterResult is List<EntityFilter>) {
      filterResult = filterResult
          .where((e) =>
              e.value != null && (e.propertyName != null || e.property != null))
          .toList();
      if (filterResult.length != 0) {
        filters = filterResult;
        loadData();
      }
    }
  }

  Future<void> onOpenMerge() async {
    var mergeResult = await AppNavigator.showAppDialog(
        context, (c) => EntityMergePage(entity: entity));
    if (mergeResult == true) {
      loadData();
    }
  }

  Future globalSearchText() async {
    if (ConnectionService.isOfflineMode && searchController.text.isNotEmpty) {
      globalSearchTextServer();
      return;
    }
    if (searchController.text.isEmpty) {
      filterdData = data;
    } else {
      filterdData = data?.where((d) {
        return searchableProperties
            .any((p) => d[p.dbName].toString().contains(searchController.text));
      }).toList();
    }

    if (entity.isTree) {
      filterdData = addChildren(filterdData, List.empty(growable: true));
    }
    setState();
  }

  Future globalSearchTextServer() async {
    if (searchController.text.isEmpty) {
      //filterdData = data;
      await loadFromServer();
      return;
    }
    if (kDebugMode) print('start search server -> ${searchController.text}');
    var result = await EntityService.getDataSearch(
      entity,
      searchController.text,
      pageIndex: pageIndex,
    );
    if (kDebugMode) print(result);
    if (!result.success) {
      result.showMessage(context);
    }

    if (result.data != null) {
      if (kDebugMode)
        print('-------------- ${result.data?.length}  records loaded ');
      filterdData = result.data;
    } else {
      filterdData = [];
    }
    if (entity.isTree) {
      filterdData = addChildren(filterdData, List.empty(growable: true));
    }
    setState();
  }

  sortBy(Property prop) async {
    if (kDebugMode) print('sort by ${prop.dbName}');
    if (prop.dbName != orderBy) {
      orderBy = prop.dbName;
      orderByDesc = false;
    } else {
      orderByDesc = !orderByDesc;
    }
    setLoading();
    await globalSearchTextServer();
    setResume();
  }

  bool _hasPageing = false;
  bool get hasPageing {
    if ((data?.length ?? 0) >= EntityService.pageSize) _hasPageing = true;
    return _hasPageing;
  }

  footerListner() async {
    if (kDebugMode)
      print('==========on load ====== ${refreshController.footerStatus}');
    if (refreshController.footerStatus != LoadStatus.loading) return;
    var success = await gotoMextPageIndex();
    if (success && hasNextPage) {
      refreshController.loadComplete();
    } else if (success && !hasNextPage) {
      //refreshController.loadComplete();
      refreshController.loadNoData();
    } else {
      refreshController.loadFailed();
    }
  }

  Future<bool> gotoMextPageIndex() {
    pageIndex++;
    pageIndexMax = max(pageIndexMax, pageIndex);
    return loadData(withLoading: false);
  }

  gotoPageIndex(int i) {
    pageIndex = i;
    setLoading();
    loadData();
  }

  Future getTotalCount() async {
    setLoading();
    if (searchController.text.isEmpty) {
      var result = await EntityService.getCount(entity,
          filter: filters ?? [], filterParentId: entity.isTree);
      if (result.data != null) {
        totalCount = result.data;
      }
    } else {
      List<EntityFilter> f = [];
      for (var prop in entity.properties.where((p) => p.isSearchable)) {
        f.add(EntityFilter(
            propertyName: prop.dbName,
            operatorId: prop.valueType == ValueTypes.Number ? 'E' : 'C',
            value: searchController.text));
      }
      var d = (await EntityService.getCount(entity,
              filter: f, filterOperator: 'or'))
          .data;
      if (d != null) totalCount = d;
    }
    setResume();
    setState();
  }

  var isLoading = false;
  Future<bool> loadData({bool withLoading = true}) async {
    if (isLoading) return true;
    isLoading = true;
    var result = await EntityService.getData(entity,
        orderBy: orderBy,
        orderByDesc: orderByDesc,
        filter: filters,
        pageIndex: pageIndex,
        filterParentId: entity.isTree);
    if (withLoading) setResume(result.success);
    if (!result.success || result.data == null) {
      //data = [];
      isLoading = false;
      if (withLoading) {
        setState();
      }
      if (ConnectionService.isOnlineMode) {
        showMessage(
          'Error Loading From Server',
          MessageType.error,
          message: result.message,
        );
      } else {
        showMessage('Error Loading From Local Database', MessageType.error);
      }
      globalSearchText();
      return false;
    } else {
      if (LayoutService.isMobile(context) && pageIndex > 0 && !entity.isLocal) {
        data?.addAll(result.data ?? []);
      } else {
        data = result.data;
      }
      if ((result.data?.length ?? 0) < EntityService.pageSize)
        hasNextPage = false;
    }
    globalSearchText();
    // if(pageIndex == 0) refreshAnimation();
    isLoading = false;
    return true;
  }

  Future loadFromServer() async {
    await loadData();

    refreshController.refreshCompleted();
  }

  Future onRefresh() async {
    pageIndex = 0;
    pageIndexMax = 0;
    hasNextPage = true;
    setLoading();
    await loadData();
  }

  Future onAdd() async {
    if (isPopup) {
      var result = await AppNavigator.showAppDialogEx(
          context, () => EntityFormPage(entity.dbName, isPopup: true),
          animationType: LayoutService.isMobile(context)
              ? AnimationTypes.slideUp
              : AnimationTypes.opacity,
          size: DialogBoxSizes.large);
      if (result == true) {
        loadFromServer();
      }
      return;
    }
    if (entityActionsService?.hasAddPage() ?? false) {
      await entityActionsService!.addPage(context);
      //if (result == true) {
      loadFromServer();
      // }
    } else {
      var result = await AppNavigator.navigateTo(
          context, () => EntityFormPage(entity.dbName),
          animationType: LayoutService.isMobile(context)
              ? AnimationTypes.slideUp
              : AnimationTypes.opacity);
      if (result == true) {
        loadFromServer();
      }
    }
  }

  Future onUpdate(d) async {
    if (isPopup) {
      var result = await AppNavigator.showAppDialogEx(
          context, () => EntityFormPage(entity.dbName, data: d, isPopup: true),
          animationType: LayoutService.isMobile(context)
              ? AnimationTypes.slideUp
              : AnimationTypes.opacity,
          size: DialogBoxSizes.large);
      if (result == true) {
        loadFromServer();
      }
      return;
    }
    if (entityActionsService?.hasEditPage(d) ?? false) {
      await entityActionsService!.editPage(context, d);
      loadFromServer();
    } else {
      var result = await AppNavigator.navigateTo(
          context, () => EntityFormPage(entity.dbName, data: d),
          animationType: LayoutService.isMobile(context)
              ? AnimationTypes.slideUp
              : AnimationTypes.opacity);
      if (result == true) {
        loadFromServer();
      }
    }
  }

  Future onView(d) async {
    if (entityActionsService?.hasEditPage(d) ?? false) {
      await entityActionsService!.editPage(context, d);
      loadFromServer();
    } else {
      var result = await AppNavigator.navigateTo(
          context, () => EntityFormPage(entity.dbName, data: d),
          animationType: LayoutService.isMobile(context)
              ? AnimationTypes.slideUp
              : AnimationTypes.opacity);
      if (result == true) {
        loadFromServer();
      }
    }
  }

  Future onDelete(d) async {
    if (!entity.permission.canDelete ||
        entity.entityType == EntityTypes.ViewOnly ||
        entity.entityType == EntityTypes.Lookup) {
      return showMessage('لا يوجد صلاحيه للحذف', MessageType.error);
    }
    if (await confirmMessage('هل انت متاكد من الحذف؟')) {
      var result = await EntityService().delete(entity, d['id']);
      if (result.success) {
        showMessage('تم الحذف بنجاح', MessageType.success);
        data?.removeWhere((e) => e['id'] == d['id']);
        globalSearchText();
        //loadFromServer();
      } else {
        showMessage(
          'خطأ اثناء الحذف',
          MessageType.error,
          message: 'الرجاء التاكد من ان العنصر غير مرتبط بعناصر اخرى',
        );
      }
    }
  }

  Future expandItem(Map<String, dynamic> d) async {
    try {
      if (!entity.isTree) return true;
      if (d['_children'] != null) {
        d['_expanded'] = !d['_expanded'];
        globalSearchText();
        return;
      }

      if (isLoading) return true;
      isLoading = true;
      if (kDebugMode) print('expand ${d['id']}');
      setLoading();
      var result = await EntityService.getData(entity,
          filter: filters,
          pageIndex: pageIndex,
          filterParentId: true,
          orderBy: orderBy,
          orderByDesc: orderByDesc,
          parentId: d['id'].toString());

      setResume();
      isLoading = false;
      if (!result.success) {
        //data = [];
        return;
      }
      for (var r in result.data ?? []) {
        if (d['_tree_level'] != null)
          r['_tree_level'] = d['_tree_level'] + 1;
        else
          r['_tree_level'] = 1;
      }
      d['_children'] = result.data;
      d['_expanded'] = true;
      // data.insertAll(
      //     data.indexWhere((i) => i['id'] == d['id']) + 1, result.data);
      globalSearchText();
    } catch (err, t) {
      print('XXX expand item error');
      print(err);
      if (kDebugMode) print(t);
      setResume();
      isLoading = false;
    }
  }

  bool hasActions(Map<String, dynamic>? d) {
    if (entity.actions.length > 0) return true;
    if (entityActionsService == null) return false;
    return (entityActionsService?.getActions(d).data?.length ?? 0) > 0;
  }

  List<EAction> getActions(Map<String, dynamic>? d) {
    return EntityService.getActions(entity, d);
  }

  Future onAction(EAction action, Map<String, dynamic> data) async {
    try {
      await EntityService.onAction(entity, action, data, context);
      onRefresh();
    } catch (err) {
      showMessage('خطأ اثناء تنفيذ الاجراء', MessageType.error);
    }
  }

  lowerFirst(String s) {
    return '${s[0].toLowerCase()}${s.substring(1)}';
  }

  String get getPhotoDataPropName =>
      '${entity.properties.firstWhere((d) => d.isImage).dbName}Id';
}
